/*****************************************************************
$Module       : XOLECTL (xolectl.cpp)
$See Also     : xolectl.h for class definition and further 
                documentation
******************************************************************/

/*****************************************************************
   Includes
******************************************************************/
#include "stdafx.h"        // Precompiled header
#include "xolectl.h"       // CExtOleControl class definition


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// Construction, Destruction, and Initialization 
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*****************************************************************
$Function:     CExtOleControl
$Description:  Creates an instance of this class
$Type:         Constructor
$Returns:      -
******************************************************************
CExtOleControl::CExtOleControl()
   :COleControl()
{
}

/*****************************************************************
$Function:     ~CExtOleControl
$Description:  Destroys an instance of this class
$Type:         Destructor
$Returns:      -
******************************************************************/
CExtOleControl::~CExtOleControl()
{
}


//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// Per-Property Browsing
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*****************************************************************
$Function:     GetEnumeratedPropertyList
$Description:  Fetches an array of enumerated property maps (i.e. 
               an array of all properties that use property maps to 
               define their valid values).
$Type:         virtual, protected
$Returns:      Pointer to first SEnumPropertyMap in the array, 
               or NULL if no array exists.

$Remarks:      A derived class must provide an implementation for 
               this.  It will be done automatically if the 
               DECLARE_PROPERTY_LIST is used in the derived class 
               definition

               The property list will be terminated with a 
               SEnumPropertyMap with a dispid of 0.
******************************************************************/
const SEnumPropertyMap* CExtOleControl::GetEnumeratedPropertyList()
{
   return NULL;
}

/*****************************************************************
$Function:     GetEnumeratedPropertyMap
$Description:  Fetches the map of possible property values for the
               specfied property dispatch id.
$Type:         virtual, protected
$Returns:      Pointer to correcponding SEnumPropertyMap

$Remarks:      This does not need to be overridden.  It will walk 
               the property list provided by the derived class 
               through GetEnumeratedPropertyList().
******************************************************************/
const SEnumPropertyMap* CExtOleControl::GetEnumeratedPropertyMap(
   DISPID dispid)
{
   // Initialize result.
   const SEnumPropertyMap* pList = NULL;

   // The the derived classes property list
   const SEnumPropertyMap* pTest = GetEnumeratedPropertyList();

   // Walk each property in the list until we find a match.
   while (pTest != NULL && pTest->m_dispid != 0 && pList==NULL)
   {
      if (pTest->m_dispid == dispid)
         pList = pTest;
      else
         pTest++;
   }

   return pList;
}

/*****************************************************************
$Function:     OnGetPredefinedStrings
$Description:  Called by the framework to obtain a set of 
               predefined strings representing the possible values 
               for a property.
$Type:         protected, virtual override
$Returns:      Nonzero if elements have been added to pStringArray 
               and pCookieArray.  0 otherwise

$Results:      This default implementation will fill the string 
               array with the values defined in the enumerated 
               property list for the given dispid.  The cookie 
               value will simply be the actual value of the 
               property.
******************************************************************/
BOOL CExtOleControl::OnGetPredefinedStrings(
   DISPID dispid, CStringArray* pStringArray, 
   CDWordArray* pCookieArray) 
{
   // Initialize result.
   BOOL bResult = FALSE;

   // Check if we have a property map for the given 
   // property dispid.
   const SEnumPropertyMap* pList = 
      GetEnumeratedPropertyMap(dispid);
   if (pList != NULL)
   {
      // Enumerate each possible value for the property.
      POSITION pos = pList->m_pMap->GetStartPosition();
      while (pos != NULL)
      {
         // Get the property value and the corresponding string 
         // table id.
         DWORD nValue(0);
         UINT  idString(0);
         pList->m_pMap->GetNextAssoc(
            pos, (void*&)nValue, (void*&)idString);

         // Add the string to the string array and the value to 
         // the cookie array.
         CString str;
         str.LoadString(idString);
         pStringArray->Add(str);
         pCookieArray->Add(nValue);
      }

      // We have a hit, return success.
      bResult = TRUE;
   }

   return bResult;
}

/*****************************************************************
$Function:     OnGetPredefinedValue
$Description:  Called by the framework to obtain the value 
               corresponding to one of the predefined strings 
               previously returned by an override of 
               COleControl::OnGetPredefinedStrings.
$Type:         protected, virtual override
$Returns:      Nonzero if a value has been returned in lpvarOut; 
               otherwise 0.

$Results:      If the property dispid is in our enumerated 
               property map, then the data value IS the value of 
               the cookie
******************************************************************/
BOOL CExtOleControl::OnGetPredefinedValue(
   DISPID dispid, DWORD dwCookie, VARIANT* lpvarOut) 
{
   BOOL bResult = FALSE;

   // Check if the property is in our map.
   const SEnumPropertyMap* pList = 
      GetEnumeratedPropertyMap(dispid);
   if (pList != NULL)
   {
      // If it's in our table, then the cookie IS the data value
      UINT idString;
      if (pList->m_pMap->Lookup(
         (void*&)dwCookie, (void*&)idString))
      {
         // assign the output data
         ::VariantInit(lpvarOut);
         lpvarOut->vt = VT_I4;
         lpvarOut->lVal = dwCookie;
      
         // Return true since we processed this.
         bResult = TRUE;
      }
   }

   return bResult;
}


/*****************************************************************
$Function:     OnGetDisplayString
$Description:  Called by the framework to obtain a string that 
               represents the current value of the property 
               identified by dispid.
$Type:         protected, virtual override
$Returns:      Nonzero if a string has been returned in strValue; 
               otherwise 0.
******************************************************************/
BOOL CExtOleControl::OnGetDisplayString(
   DISPID dispid, CString& strValue) 
{
   BOOL bResult = FALSE;

   // Check if the property is in our map.
   const SEnumPropertyMap* pList = 
      GetEnumeratedPropertyMap(dispid);
   if (pList != NULL)
   {
      // Make sure this property has saved the address of it's
      // member variable
      if (pList->m_numBytes == sizeof(short) || 
         pList->m_numBytes == sizeof(long))
      {
         // Read the current value from the property
         long nValue = -1;
         if (pList->m_numBytes == sizeof(short))
            nValue = *(short*)(((char*)this) + pList->m_nOffset);
         else if (pList->m_numBytes == sizeof(long))
            nValue = *(long*)(((char*)this) + pList->m_nOffset);

         // Lookup the corresponding string for  this value in 
         // the map
         UINT idString;
         if (pList->m_pMap->Lookup(
            (void*&)nValue, (void*&)idString))
         {
            strValue.LoadString(idString);
      
            // Return true since we processed this.
            bResult = TRUE;
         }
      }
   }

   return bResult;
}

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// Internal Property Enumeration classes
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

/*****************************************************************
$Function:     SEnumPropertyValue::SEnumPropertyValue
$Description:  Creates and initializes an instance of this class
$Type:         Constructor
$Returns:      -
******************************************************************/
SEnumPropertyValue::SEnumPropertyValue(long value, UINT idString)
   :m_nValue(value)
   ,m_idString(idString)
{
}

/*****************************************************************
$Function:     SEnumPropertyMap::SEnumPropertyMap
$Description:  Creates and initializes an instance of this class
$Type:         Constructor
$Returns:      -
******************************************************************/
SEnumPropertyMap::SEnumPropertyMap
(
   DISPID dispid,    // (I) dispatch id of the property to map 
                     //     values for
   long   nOffset,
   long   numBytes,
   ...               // (I) Variable list of SEnumPropertyValue 
                     //  objects, terminated with a property with 
                     //  value and string id of 0.
)
   :m_dispid(dispid)
   ,m_nOffset(nOffset)
   ,m_numBytes(numBytes)
   ,m_pMap(NULL)
{
   // Allocate a map
   m_pMap = new CMapPtrToPtr;

   // Make sure we have a valid dispid
   if (m_dispid != 0)
   {
      // Initialize the variable argument list
      va_list args;
      va_start(args, numBytes);

      // Walk through the variable list of property values, 
      // adding each one to the map
      SEnumPropertyValue* pEnum = 
         va_arg(args, SEnumPropertyValue*);

      while (pEnum!=NULL && pEnum->m_idString!=0)
      {
         m_pMap->SetAt(
            (void*)pEnum->m_nValue, (void*)pEnum->m_idString);

         // save this property value
         SEnumPropertyValue* pEnumVictem = pEnum;

         // get the next property value
         pEnum = va_arg(args, SEnumPropertyValue*);

         // Free the previous property value.  The objects passed 
         // to us are newed by yhe PROPERTY_VALUE macro.  We must 
         // free it here. Passing stack objects was not working, 
         // so we had to new them.
         delete pEnumVictem;
      }

      // terminate the variable argument parsing
      va_end(args);
   }
}

/*****************************************************************
$Function:     SEnumPropertyMap::~SEnumPropertyMap
$Description:  Destroys an instance of this class
$Type:         Destructor
$Returns:      -
******************************************************************/
SEnumPropertyMap::~SEnumPropertyMap()
{
   // Free the property map
   if (m_pMap != NULL)
      delete m_pMap;
}
